#include <math.h>
#include <stdlib.h>
#include "nrutil.c"
#include "rkck.c"


float **y, *xx; 

float M1, M2, L0, L1, L2, L3, D1, D2, D3;


void rk4 (float y[], 
		  float dydx[], 
		  int n, 
		  float x, 
		  float h, 
		  float yout[],
		  void (*derivs) (float, 
		                  float [], 
						  float [])
		  )
/* 
Given values for the variables y[1..n] and their derivatives dydx[1..n] known at x, use the
fourth-order Runge-Kutta method to advance the solution over an interval h and return the
incremented variables as yout[1..n], which need not be a distinct array from y. The user
supplies the routine derivs(x,y,dydx), which returns derivatives dydx at x. 
*/
{
	int i;
	float xh,hh,h6,*dym,*dyt,*yt;
	
	dym=vector(1,n);
	dyt=vector(1,n);
	yt=vector(1,n);

	hh=h*0.5;
	h6=h/6.0;
	xh=x+hh;
	
	for (i=1;i<=n;i++) yt[i]=y[i]+hh*dydx[i]; // First step.
	
	(*derivs)(xh,yt,dyt);                     // Second step.
	for (i=1;i<=n;i++) yt[i]=y[i]+hh*dyt[i];

	(*derivs)(xh,yt,dym);                     // Third step.
	for (i=1;i<=n;i++) {
		yt[i]=y[i]+h*dym[i];
		dym[i] += dyt[i];
	}

	(*derivs)(x+h,yt,dyt);                    // Fourth step.
	for (i=1;i<=n;i++)                        // Accumulate increments with proper weights. 
		yout[i]=y[i]+h6*(dydx[i]+dyt[i]+2.0*dym[i]);

	free_vector(yt,1,n);
	free_vector(dyt,1,n);
	free_vector(dym,1,n);
}



void rkdumb (float vstart[], 
			 int nvar, 
			 float x1, 
			 float x2, 
			 int nstep,
			 void (*derivs)(float, 
							float [], 
							float [])
			 )
/*
Starting from initial values vstart[1..nvar] known at x1 use fourth-order Runge-Kutta
to advance nstep equal increments to x2. The user-supplied routine derivs(x,v,dvdx)
evaluates derivatives. Results are stored in the global variables y[1..nvar][1..nstep+1]
and xx[1..nstep+1].
*/
{
	void rk4 (float y[], float dydx[], int n, float x, float h, float yout[],
			  void (*derivs)(float, float [], float []));
	int i,k;
	float x,h;
	float *v,*vout,*dv;

	v=vector(1,nvar);
	vout=vector(1,nvar);
	dv=vector(1,nvar);

	for (i=1;i<=nvar;i++) { // Load starting values.
		v[i]=vstart[i];
		y[i][1]=v[i];
	}

	xx[1]=x1;
	x=x1;
	h=(x2-x1)/nstep;
	
	for (k=1;k<=nstep;k++) { // Take nstep steps.
	
		(*derivs)(x,v,dv);
		rk4(v,dv,nvar,x,h,vout,derivs);
		
		if ((float)(x+h) == x) nrerror("Step size too small in routine rkdumb");
		
		x += h;
		xx[k+1]=x;			// Store intermediate steps.
		
		for (i=1;i<=nvar;i++) {
			v[i]=vout[i];
			y[i][k+1]=v[i];
		}
	}
	
	free_vector(dv,1,nvar);
	free_vector(vout,1,nvar);
	free_vector(v,1,nvar);
}



void euler_step (int N,
				 float x, 
	 			 float y[], 
				 float h, 
				 void (*f) (float, float [], float []), 
				 float ynext[])
{
	float *dyperdx;
	int n;

	dyperdx = vector (1, N);

	(*f) (x, y, dyperdx);

	for (n=1; n<=N; n++)	
	{
		ynext[n] = y[n] + h*dyperdx[n];  
	}

	free_vector (dyperdx, 1, N);
}


void euler_drv (float x0, 
				float x1, 
				int xstep,
				int N,
				float y0[],
				void (*f) (float, float [], float []))
{
	float h, *yn, *yn1;
	int i, n;

	h = (x1-x0)/xstep;

	yn = vector (1, N);
	yn1 = vector (1, N);

	xx[1] = x0;

	for (n=1; n<=N; n++) 
	{	
		y[n][1] = y0[n];
		yn[n] = y0[n];
	}

	for (i=1; i<=xstep; i++)
	{
		euler_step (N,
					xx[i], 
					yn, 
					h, 
					f, 
					yn1);

		xx[i+1] = xx[i]+h;
		
		for (n=1; n<=N; n++) 
		{	
			y[n][i+1] = yn1[n];
			yn[n] = yn1[n];
		}
	}

	free_vector (yn, 1, N);
	free_vector (yn1, 1, N);

}

void f_csatolt_rezges (float x, float y[], float dyperdx[])
{
										// y[1]: x1(t), y[2]: x2(t), y[3]: v1(t), y[4]: v2(t)
		dyperdx[1] = y[3];
		dyperdx[2] = y[4];
	
	dyperdx[3] = (-D1*(y[1]-L1) + D3*(y[2]-y[1]-L3))/M1;	
	dyperdx[4] = (-D2*(y[2]-(L0-L2)) - D3*(y[2]-y[1]-L3))/M2;
}


void main (int argc, char *argv[])
{

	int i, xstep, n, N, method, nok, nbad;
	float x0, x1, *y0, nrg, eps, sum;


	x0 = 0.0;
	x1 = 1.0;
	xstep = 1024;

	N = 4;
	y0 = vector (1, N);			// kezdeti ertekek es alapertelmezett ertekei
	y0[1] = 1.0;
	y0[2] = 2.0;
	y0[3] = 0;
	y0[4] = 0;

	for (i=1; i<argc; i++ ){
		
		switch(i){
	    case 1: 
			x0 = atof(argv[i]); 
			break;
		case 2:
			x1 = atof(argv[i]); 
			break;
		case 3:
			xstep = atoi(argv[i]); 
			break;
		case 4:
			y0[1] = atof(argv[i]);
			break;
		case 5:
			y0[2] = atof(argv[i]);
			break;
		case 6:
			y0[3] = atof(argv[i]);
			break;
		case 7:
			y0[4] = atof(argv[i]);
			break;
		case 8:
			M1 = atof(argv[i]);
			break;
		case 9:
			M2 = atof(argv[i]);
			break;
		case 10:
			L0 = atof(argv[i]);
			break;
		case 11:
			L1 = atof(argv[i]);
			break;
		case 12:
			L2 = atof(argv[i]);
			break;
		case 13:
			L3 = atof(argv[i]);
			break;
		case 14:
			D1 = atof(argv[i]);
			break;
		case 15:
			D2 = atof(argv[i]);
			break;
		case 16:
			D3 = atof(argv[i]);
			break;
		case 17:
			method = atoi(argv[i]);
			break;
		} 
	}


	switch(method)
	{
	 case 1: 
		
		 xx = vector (1, xstep+1);
		 y = matrix (1, N, 1, xstep+1);

		euler_drv (	x0, 
					x1, 
					xstep,
					N,
					y0,
					f_csatolt_rezges);

		for (i=1;i<=xstep+1;i++)
		{
			nrg = D1*(y[1][i]-L1)*(y[1][i]-L1) 
				+ D3*(y[2][i]-y[1][i]-L3)*(y[2][i]-y[1][i]-L3) 
				+ D2*(y[2][i]-(L0-L2))*(y[2][i]-(L0-L2))
				+ M1*y[3][i]*y[3][i]
				+ M2*y[4][i]*y[4][i];
			
			printf("%f", xx[i]);
			
			for (n=1; n<=N; n++) 
			{	
				printf("\t%f", y[n][i]);
			}

			printf("\t%f\n", nrg);
			
		}

		free_vector (xx, 1, xstep+1);
		free_vector (y0, 1, N);
		free_matrix (y, 1, N, 1, xstep+1);
		
		break;
	 
	 case 2:

		 xx = vector (1, xstep+1);
		 y = matrix (1, N, 1, xstep+1);

		
		 rkdumb (y0, 
				N, 
				x0, 
				x1, 
				xstep,
				f_csatolt_rezges);

		for (i=1;i<=xstep+1;i++)
		{
			nrg = D1*(y[1][i]-L1)*(y[1][i]-L1) 
				+ D3*(y[2][i]-y[1][i]-L3)*(y[2][i]-y[1][i]-L3) 
				+ D2*(y[2][i]-(L0-L2))*(y[2][i]-(L0-L2))
				+ M1*y[3][i]*y[3][i]
				+ M2*y[4][i]*y[4][i];
			
			printf("%f", xx[i]);
			
			for (n=1; n<=N; n++) 
			{	
				printf("\t%f", y[n][i]);

			}
			printf("\t%f\n", nrg);
		}
		free_vector (xx, 1, xstep+1);
		free_vector (y0, 1, N);
		free_matrix (y, 1, N, 1, xstep+1);
		
		break;

	 case 3:
		 
		 kmax = 10000;
		 xp = vector (1, kmax);
		 yp = matrix (1, N, 1, kmax);
		 dxsav = (x1-x0)/1000; 
		 eps = 0.00000001;

		 hdone = vector (1, MAXSTP);
		 hwill = vector (1, MAXSTP);
		 for (i=1;i<=MAXSTP;i++)
		 {
			 hdone[i] = 0.0;
			 hwill[i] = 0.0;
		 }
		 
		 odeint(y0, 
				N, 
				x0, 
				x1, 
				eps, 
				(x1-x0)/xstep,
				0, 
				&nok, 
				&nbad,
				f_csatolt_rezges,
				rkqs);

		 for (i=1;i<=kount;i++)
		 {
			nrg = D1*(yp[1][i]-L1)*(yp[1][i]-L1) 
				+ D3*(yp[2][i]-yp[1][i]-L3)*(yp[2][i]-yp[1][i]-L3) 
				+ D2*(yp[2][i]-(L0-L2))*(yp[2][i]-(L0-L2))
				+ M1*yp[3][i]*yp[3][i]
				+ M2*yp[4][i]*yp[4][i];
			
			printf("%f", xp[i]);
			
			for (n=1; n<=N; n++) 
			{	
				printf("\t%f", yp[n][i]);

			}
			printf("\t%f\n", nrg);
		 }
		 
		 sum = 0.0;
		 for (i=1;i<=MAXSTP;i++)
		 {
			 sum += hdone[i];
			 printf("%f\t%f\t%f\n", sum, hdone[i], hwill[i]);
		 }

		 free_vector (xp, 1, kmax);
		 free_vector (y0, 1, N);
		 free_matrix (yp, 1, N, 1, kmax);
		 free_vector (hdone, 1, MAXSTP);
		 free_vector (hwill, 1, MAXSTP);

		 break;
	 
	}


}
